# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

ORGANIZATION = 'organization'
MODE_ACCESS_TOKEN = 'mode_user_token'
MODE_PASSWORD_TOKEN = 'mode_password_token'
